#ifndef HYPER_GENOM_GENOM_TCLSERV_CLIENT_MODULE_HH_
#define HYPER_GENOM_GENOM_TCLSERV_CLIENT_MODULE_HH_

#include <map>

#include "tclserv_client_queue.hh"

namespace hyper {
	namespace genom_model {

	template <typename Input, typename Output, typename Error>
	struct method;

	template <typename Handler, typename Error>
	void handle_error(int err, Error bilan, boost::tuple<Handler> handler)
	{
		using namespace boost::system;
		switch (err) {
			case -2:
				return boost::get<0>(handler)(make_error_code(errc::protocol_error));
			case -1:
				return boost::get<0>(handler)(make_error_code(bilan));
			case  0:
				return boost::get<0>(handler)(error_code());
		}
	}

	template <typename Error>
	struct method<void, void, Error>
	{
		typedef int (*TFuncIn)(TCLSERV_CLIENT_ID, ssize_t*);
		typedef int (*TFuncOut)(TCLSERV_CLIENT_ID, ssize_t, int*);

		TFuncIn FuncIn;
		TFuncOut FuncOut;
		tclserv_client_queue& queue;

		method(TFuncIn In, TFuncOut Out, tclserv_client_queue& queue) : 
			FuncIn(In), FuncOut(Out), queue(queue) {}

		template <typename Handler>
		void handle_request(ssize_t rqstId, boost::tuple<Handler> handler)
		{
			int bilan;
			int err = FuncOut(queue.get_client_id(), rqstId, &bilan);
			handle_error(err, static_cast<Error>(bilan), handler);
		}
		
		template <typename Handler>
		ssize_t operator() (Handler handler) 
		{
			using namespace boost::system;
			ssize_t rqstId;
			int err = FuncIn(queue.get_client_id(), &rqstId); 
			if (err < 0) {
				handler(error_code(errno, system_category()));
				return -1;
			}
			
			void (method::*f) (ssize_t, boost::tuple<Handler> handler) = 
				&method::template handle_request<Handler>;

			queue.enqueue(rqstId, boost::bind(f, this, rqstId, boost::make_tuple(handler)));
			return rqstId;
		}
	};

	template <typename Input, typename Error>
	struct method<Input, void, Error>
	{
		typedef int (*TFuncIn)(TCLSERV_CLIENT_ID, ssize_t*, Input*);
		typedef int (*TFuncOut)(TCLSERV_CLIENT_ID, ssize_t, int*);

		TFuncIn FuncIn;
		TFuncOut FuncOut;
		tclserv_client_queue& queue;

		method(TFuncIn In, TFuncOut Out, tclserv_client_queue& queue) : 
			FuncIn(In), FuncOut(Out), queue(queue) {}

		template <typename Handler>
		void handle_request(ssize_t rqstId, boost::tuple<Handler> handler)
		{
			int bilan;
			int err = FuncOut(queue.get_client_id(), rqstId, &bilan);
			handle_error(err, static_cast<Error>(bilan), handler);
		}
	
		template <typename Handler>
		ssize_t operator() (const Input& input, Handler handler) 
		{
			using namespace boost::system;
			ssize_t rqstId;
			int err = FuncIn(queue.get_client_id(), &rqstId, const_cast<Input*>(&input)); 
			if (err < 0) {
				handler(error_code(errno, system_category()));
				return -1;
			}
			
			void (method::*f) (ssize_t, boost::tuple<Handler> handler) = 
				&method::template handle_request<Handler>;

			queue.enqueue(rqstId, boost::bind(f, this, rqstId, boost::make_tuple(handler)));
			return rqstId;
		}
	};

	template <typename Output, typename Error>
	struct method<void, Output, Error>
	{
		typedef int (*TFuncIn)(TCLSERV_CLIENT_ID, ssize_t*);
		typedef int (*TFuncOut)(TCLSERV_CLIENT_ID, ssize_t, Output*, int*);

		TFuncIn FuncIn;
		TFuncOut FuncOut;
		tclserv_client_queue& queue;

		method(TFuncIn In, TFuncOut Out, tclserv_client_queue& queue) : 
			FuncIn(In), FuncOut(Out), queue(queue) {}

		template <typename Handler>
		void handle_request(ssize_t rqstId, Output& output, boost::tuple<Handler> handler)
		{
			int bilan;
			int err = FuncOut(queue.get_client_id(), rqstId, &output, &bilan);
			handle_error(err, static_cast<Error>(bilan), handler);
		}
	
		
		template <typename Handler>
		ssize_t operator() (Output& output, Handler handler) 
		{
			using namespace boost::system;
			ssize_t rqstId;
			int err = FuncIn(queue.get_client_id(), &rqstId); 
			if (err < 0) {
				handler(error_code(errno, system_category()));
				return -1;
			}
			
			void (method::*f) (ssize_t, Output& output, boost::tuple<Handler> handler) = 
				&method::template handle_request<Handler>;

			queue.enqueue(rqstId, 
						  boost::bind(f, this, rqstId, boost::ref(output), boost::make_tuple(handler)));
			return rqstId;
		}
	};

	template <typename Input, typename Output, typename Error>
	struct method 
	{
		typedef int (*TFuncIn)(TCLSERV_CLIENT_ID, ssize_t*, Input*);
		typedef int (*TFuncOut)(TCLSERV_CLIENT_ID, ssize_t, Output*, int*);

		TFuncIn FuncIn;
		TFuncOut FuncOut;
		tclserv_client_queue& queue;

		method(TFuncIn In, TFuncOut Out, tclserv_client_queue& queue) : 
			FuncIn(In), FuncOut(Out), queue(queue) {}

		template <typename Handler>
		void handle_request(ssize_t rqstId, Output& output, boost::tuple<Handler> handler)
		{
			int bilan;
			int err = FuncOut(queue.get_client_id(), rqstId, &output, &bilan);
			handle_error(err, static_cast<Error>(bilan), handler);
		}
	
		
		template <typename Handler>
		ssize_t operator() (const Input& input, Output& output, Handler handler) 
		{
			using namespace boost::system;
			ssize_t rqstId;
			int err = FuncIn(queue.get_client_id(), &rqstId, const_cast<Input*>(&input)); 
			if (err < 0) {
				handler(error_code(errno, system_category()));
				return -1;
			}
			
			void (method::*f) (ssize_t, Output& output, boost::tuple<Handler> handler) = 
				&method::template handle_request<Handler>;

			queue.enqueue(rqstId, 
						  boost::bind(f, this, rqstId, boost::ref(output), boost::make_tuple(handler)));
			return rqstId;
		}
	};
	}
}

#endif
